#!/bin/bash

#
# Copyright (C) 2015 SpinetiX
#

NAME=bootspx.sh

#
# Assorted initialization things, run late in rcS.
# It should run after bootmisc.sh
#

# Sets the system time precisely from the hardware clock (RTC),
# normally this is initialized loosely by the kernel (with a 1 sec
# precision).
set_systime_from_hc() {
    hwclock --utc --noadjfile --hctosys
    if [ $? -ne 0 ]; then
	echo 'failed to precisely set system time from hardware clock'
    fi
}

set_fallback_systime() {
    if [ ! -f /etc/os-release ]; then
	echo "missing /etc/os-release, cannot check system time"
	return
    fi
    reftime=$(date -u -r /etc/os-release +%4Y%2m%2d%2H%2M)
    systime=$(date -u +%4Y%2m%2d%2H%2M)
    if [ "$reftime" -gt "$systime" ]; then
	echo 'system time older than firmware creation time, setting to latter'
	date -u ${reftime#????}${systime%????????}
    fi
}

# Apply extra sysctl from /etc/sysctl.d, base /etc/sysctl.conf is
# already applied by bootmisc.sh
apply_extra_sysctl() {
    for file in /etc/sysctl.d/*.conf; do
	sysctl -p "$file"
    done
}

# applies all net.<proto>.conf.default sysctl settings to the already
# initialized interfaces; if an interface was already up (e.g.,
# NFS-root) before the net.<proto>.conf.default setting was applied
# then it would not have picked up the change, this propagates these
# settings to them
fixup_iface_sysctl() {
    # The sed command below, input to read loop, gets all
    # net.<proto>.conf.default.<parameter>=<value> in sysctl.conf
    # files and reports three space separated fields: the
    # net/<proto>/conf, <parameter> and <value>.
    #
    # Note that the read loop executes in a subshell, so it does not
    # affect this subshell's variables.
    sed -n -e '/^\s*net\.[^\s.]*\.conf\.default\.\S\+\s*=/{s:.default.: :;s:\.:/:g;s:=: :;p}' \
	/etc/sysctl.conf /etc/sysctl.d/*.conf 2>/dev/null | \
	while read -r path param value; do
	    procpath=/proc/sys/$path
	    [ -e "$procpath" ] || continue
	    # apply to all interfaces (except pseudo "default" and "all")
	    for file in $procpath/*/$param; do
		[ "$file" = "$procpath/default/$param" ] && continue
		[ "$file" = "$procpath/all/$param" ] && continue
		echo "$value" > "$file"
	    done
        done
}

# acquires a lock for the serial port on which the console is running;
# the lock is acquired for init's PID (1) since it as long lived as
# the kernel. It also sets the closing_wait parameter to 100 (i.e. 1
# second) so that if the serial port is suspended due to flow control
# it does not hang the shutdown process (default is 30 seconds).
lock_serial_console() {
    local port

    port="$(sed -e 's/.*\bconsole=\(tty[^[:digit:][:space:]][[:alnum:]]*\).*/\1/p;d' < /proc/cmdline)"
    [ -z "$port" ] && return

    echo -n "Registering $port for system console: lock"
    lockdev -l -p1 "$port" || echo -n " failed"
    case "$port" in
	ttyO*)
	    # OMAP serial ports fail with setserial (it returns "invalid argument")
	    ;;
	*)
	    echo -n ", "
	    echo -n "port settings"
	    setserial /dev/"$port" closing_wait 100 || echo -n " failed"
	    ;;
    esac
    echo "."
}

# verifies that if a ssh host key exists then the public key pair
# exists and both have non-zero size, if the check fails then the keys
# are removed so that they are generated when sshd starts.
verify_ssh_keys() {
    local kf
    for kf in /etc/ssh/ssh_host_*_key ; do
	[ -f "$kf" ] || continue
	[ -s "$kf" -a -s "$kf".pub ] && continue
	rm -f "$kf" "$kf".pub
    done
}

case "$1" in
    start|restart)
	set_systime_from_hc
	set_fallback_systime
	apply_extra_sysctl
	fixup_iface_sysctl
	lock_serial_console
	verify_ssh_keys
	;;
    stop|reload|force-reload)
	;;
    *)
esac
    
	
