#!/bin/bash
#
# License: Copyright 2015 SpinetiX S.A. This file is licensed
#          under the terms of the GNU General Public License version 2.
#          This program is licensed "as is" without any warranty of any
#          kind, whether express or implied.
#
### BEGIN INIT INFO
# Required-Start:
# Required-Stop:
# Should-Start:
# Should-Stop:
# Default-Start: S
# Default-Stop: 0 1 2 3 4 5 6
# Short-Description: Populate volatile and temporary files and directories
# Description: Populate volatile and temporary files and directories
### END INIT INFO

# Init script information
NAME=spxtmpfiles.sh

# The directory where to find the tmpfile.d spec
#
# NOTE: the freedesktop.org specification has /etc/tmpfiles.d,
# /run/tmpfiles.d and /usr/lib/tmpfiles.d as possible locations, in
# order of precedence (*.conf files with the same name in the first
# location overriding others). Packages should install their files in
# /etc/tmpfiles.d. We do not implement the other locations.
TMPFILESCONFDIR=/etc/tmpfiles.d

# Source the init script functions
. /etc/init.d/functions

set_path_attrs() {
    local path="$1"
    local mode="$2"
    local uid="$3"
    local gid="$4"

    [ -n "$mode" -a "$mode" != "-" ] && chmod -- "$mode" "$path"
    [ -n "$uid" -a "$uid" != "-" ] && chown -- "$uid" "$path"
    [ -n "$gid" -a "$gid" != "-" ] && chgrp -- "$gid" "$path"
}

# Parses the tmpfiles specification (compatible with systemd) given as
# the first arg and performs all the required actions. The type
# identifiers handled are those of an old version of the
# specification, newer versions introduce extra types not handled
# here.
do_tmpfiles_conf() {
    local conf="$1"
    local type path mode uid gid junk
    while read type path mode uid gid junk; do
	[ -z "$type" -o "$type" = "#" ] && continue # empty line or comment
	if [ -z "$path" ]; then
	    echo "empty path in tmpfiles spec" >&2
	    continue
	fi
	if [ "${path:0:1}" != "/" ]; then
	    echo "relative path ('$path') in tmpfiles spec not allowed" >&2
	    continue
	fi
	case "$type" in
	    f) # create a file if it does not exist yet
		if [ ! -e "$path" ]; then
		    echo -n > "$path"
		    set_path_attrs "$path" "$mode" "$uid" "$gid"
		fi
		;;
	    F) # create or truncate a file
		echo -n > "$path"
		set_path_attrs "$path" "$mode" "$uid" "$gid"
		;;
	    d) # create a directory if it does not exist yet
		if [ ! -e "$path" ]; then
		    mkdir -- "$path"
		    set_path_attrs "$path" "$mode" "$uid" "$gid"
		fi
		;;
	    D) # create or empty a directory
		if [ ! -e "$path" ]; then
		    mkdir -- "$path"
		    set_path_attrs "$path" "$mode" "$uid" "$gid"
		else
		    find -- "$path" -mindepth 1 -print0 | xargs -0r rm -rf --
		fi		    
		;;
	    x) # ignore a path
		;;
	    r) # remove a path
		if [ -d "$path" ]; then
		    rmdir -- "$path"
		elif [ -e "$path" ]; then
		    rm -f -- "$path"
		fi
		;;
	    R) # recursively remove a path
		[ -e "$path" ] && rm -rf -- "$path"
		;;
	    *) # unkown specification
		echo "invalid tmpfiles spec '$type'" >&2
		;;
	esac
    done < "$conf"
}

# Do all the tmpfiles specifications found
do_tmpfiles_all() {
    for f in "$TMPFILESCONFDIR"/*.conf ; do
	[ -f "$f" ] && do_tmpfiles_conf "$f"
    done
}

#
# Main
#

case "$1" in
    start)
	echo -n "Populating volatile and temporary paths: $NAME "

	do_tmpfiles_all

	success
	echo

	;;
    stop)
	;;
esac
